import json
from abc import ABC, abstractmethod
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.model_selection import train_test_split
from sklearn.naive_bayes import MultinomialNB
from sklearn import metrics
import copy

class Intent(ABC):

    @abstractmethod
    def __init__(self, intentName="Irrelevant",allIntentExamples=None, entityHandler=None,testSize=0.25,vocab=None):

        self.intentName = intentName
        self.testSize = testSize

        if not allIntentExamples:
            allIntentExamples = {}

        self.entityHandler = entityHandler
        self.allIntentExamples = allIntentExamples
        if self.entityHandler:
            self.substituteAllEntitiesInAllTrainingExamples()

        self.classifier = None
        self.vocab=vocab

        self.trainedIntentArray = []
        self.trained = False
        self.createIntent()

    def substituteAllEntitiesInAllTrainingExamples(self):

        """

        Substitutes all the training examples with the entity classes

        :return:
        """

        allIntentExamplesSubstituted = {}
        allIntentExamplesCopy = copy.deepcopy(self.allIntentExamples)

        for key in allIntentExamplesCopy:
            dum = []
            for mes in allIntentExamplesCopy[key]:
                dum.append(self.entityHandler.substituteAllEntities(" {0} ".format(mes)))

            allIntentExamplesSubstituted[key] = dum

        self.allIntentExamples = allIntentExamplesSubstituted

    @abstractmethod
    def createIntent(self,clfName=None):
        pass

    @staticmethod
    @abstractmethod
    def train(X_train, y_train, params = None):
        pass


class MultinomialNBIntent(Intent):

    def __init__(self,intentName="Irrelevant",allIntentExamples=None,entityHandler = None,vocab=None):
        """

        :param allIntentExamples:
        :param entityHandler:
        """

        super().__init__(intentName=intentName,allIntentExamples=allIntentExamples,
                         entityHandler = entityHandler,testSize=0.25,vocab=vocab)


    def substituteAllEntitiesInAllTrainingExamples(self):
        """

        :return:
        """
        super().substituteAllEntitiesInAllTrainingExamples()


    def createVocab(self):

        ##
        ##
        ## this function creates the vocabulary for the BOW /
        ##
        ##
        ##
        ## Initialise the Count Vectoriser
        ##
        count_vect = CountVectorizer() # stop_words="english"
        ##
        ## Create the Count Vectoriser Object (This gives a bag of words representation)
        ##
        allIntentExamplesCopy = copy.deepcopy(self.allIntentExamples)

        allData=[intentExample for intentExamples in allIntentExamplesCopy.values() for intentExample in intentExamples]
        count_vect.fit_transform(allData)
        vocab=count_vect.vocabulary_
        count_vect.revVocabulary_ =  dict(zip(vocab.values(),vocab.keys()))
        self.vocab = count_vect

        return count_vect

    def createDataSetForIntent(self):

        ##
        ## if output is 1 means intent is present else not
        ##
        allIntentExamplesCopy = copy.deepcopy(self.allIntentExamples)

        intentSamples = allIntentExamplesCopy[self.intentName]
        #
        # pop the intent from json
        #
        allIntentExamplesCopy.pop(self.intentName, None)
        nonIntentSamples = [intentExample for intentExamples in allIntentExamplesCopy.values() for intentExample in
                            intentExamples]
        Y = [1 for i in range(len(intentSamples))] + [0 for i in range(len(nonIntentSamples))]
        intentSamples.extend(nonIntentSamples)
        X = copy.deepcopy(intentSamples)

        ##
        ## stratify split the dataset to maintain class balance
        ##

        X_train, X_test, y_train, y_test = train_test_split(X, Y, test_size=self.testSize, stratify=Y, random_state=2020)

        return X_train, X_test, X, y_train, y_test, Y

    @staticmethod
    def train(X_train, y_train, params = None):
        vectorizer = params["vectorizer"]
        X_trainCount = vectorizer.transform(X_train)
        clf = MultinomialNB()
        clf.fit(X_trainCount, y_train)
        return clf

    @staticmethod
    def predictClassifierOutput(clf, X, count_vect):
        X_count = count_vect.transform(X)
        predictions = clf.predict(X_count)
        probabilities = clf.predict_proba(X_count)
        return predictions, probabilities


    def getIntentProbability(self, userMessage):
        userMessage = [userMessage]
        predictions, probabilities = self.predictClassifierOutput(self.classifier["classifier"], userMessage,
                                                             self.classifier["data"]["countVect"])

        # print probabilities
        return {"name": self.classifier["name"], "confidence": probabilities[0][1]}

    def createIntent(self, clfName=None):
        if not clfName:
            clfName = self.intentName + "_multinomialNbClassifier"

        ##
        ## Create vocabulary
        ##

        if not self.vocab:
            cv = self.createVocab()
        else:
            cv= self.vocab

        X_train, X_test, X_all, y_train, y_test, y_all = self.createDataSetForIntent()

        clf = self.train(X_train, y_train, {"vectorizer" :cv})

        predictions, probabilities = self.predictClassifierOutput(clf, X_test, cv)

        expectedAccuracy = metrics.accuracy_score(y_test, predictions)

        ##
        ## Now train on the complete dataset
        ##
        clf = self.train(X_all, y_all, {"vectorizer" :cv})

        classifier = {
            "name": clfName,
            "classifier": clf,
            "expectedAccuracy": expectedAccuracy,
            "data": {
                "countVect": cv,
                "trainedOnX": X_train,
                "trainedOnY": y_train
            }
        }

        self.classifier = classifier
        return classifier



class IntentsHandler():

    def __init__(self,allIntentExamples=None, entityHandler=None):

        self.allIntentExamples = allIntentExamples
        self.intentSamplesAugment()
        self.entityHandler     = entityHandler
        self.trained    = False
        self.createAllTrainedIntents()


    def intentSamplesAugment(self):
        # Write a better function for this
        for i in range(4):
            for key in self.allIntentExamples:
                self.allIntentExamples[key].extend(self.allIntentExamples[key])
                if len(self.allIntentExamples[key]) < 20:
                    self.allIntentExamples[key].extend(self.allIntentExamples[key])


    def createAllTrainedIntents(self):

        allIntentExamplesCopy = copy.deepcopy(self.allIntentExamples)

        trainedIntentArray = []
        intentNames  = list(allIntentExamplesCopy.keys())
        intentNames.sort()
        #trainedIntentArray.append(MultinomialNBIntent(intentName=intentNames[0],allIntentExamples=self.allIntentExamples,entityHandler=self.entityHandler))
        #vocab = trainedIntentArray[0].vocab

        for intentName in intentNames:
            #
            # allIntentExamplesCopy = copy.deepcopy(self.allIntentExamples)
            # print intent
            #
            trainedIntentArray.append(
                MultinomialNBIntent(intentName=intentName,allIntentExamples=self.allIntentExamples,entityHandler=self.entityHandler,vocab=None)

            )
        self.trainedIntentArray =trainedIntentArray
        self.trained=True



    def getMessageIntents(self, message,dialogNumber=0):

        if self.entityHandler:
            message = self.entityHandler.substituteAllEntities(message)

        if self.trained:
            allIntents=sorted([trainedIntent.getIntentProbability(message) for trainedIntent in self.trainedIntentArray],
                      key=lambda x: x["confidence"], reverse=True)

            allIntentsDialogCounterUpdated=[]
            rank=1
            for elem in allIntents:
                elem["dialogNumber"]=dialogNumber
                elem["rank"]=rank
                allIntentsDialogCounterUpdated.append(elem)
                rank+=1

            return allIntentsDialogCounterUpdated
        else:
            return "error"






from .Intent import IntentsHandler
from .Entity import EntitiesHandler
from objectpath import Tree
import datetime

class ContextManager():

    def __init__(self,entitiesExtractorJson,allIntentExamples):

        entitiesHandler =  EntitiesHandler(entitiesExtractorJson)
        intentHandlers  = [IntentsHandler(allIntentExamples,entityHandler=entitiesHandler)]

        self.entitiesHandler = entitiesHandler
        self.intentHandlers  = intentHandlers
        self.context={
            "entities":[],
            "intents":[],
            "dialogs":[],
            "messages":[]
        }
        self.contextTree = Tree(self.context)

    def setNewContext(self,newContext):
        self.context = newContext
        self.contextTree =Tree(self.context)

    def clearContext(self):
        context = {
            "entities": [],
            "intents": [],
            "dialogs": [],
            "messages": []
        }
        self.setNewContext(context)



    def updateDialog(self,message,type="input"):

        if type =="input":
            # for input
            dialogNumber = len(self.context["dialogs"]) + 1
            extractedEntities = self.entitiesHandler.extractAllEntities(message,dialogNumber=dialogNumber)
            extractedIntents  = [ intent for ihandler in self.intentHandlers for intent in ihandler.getMessageIntents(message=message,dialogNumber=dialogNumber) ]

            self.context["entities"].extend(extractedEntities)
            self.context["intents"].extend(extractedIntents)
            self.context["dialogs"].append(dialogNumber)
            messageJson={
                "text":message,
                "entitiesSub":self.entitiesHandler.substituteAllEntities(message),
                "messageType":type,
                "dialogNumber":dialogNumber,
                "sendTime":  datetime.datetime.now().strftime("%d/%m/%Y, %H:%M:%S")

            }
        else:
            # for output message
            dialogNumber = len(self.context["dialogs"])
            messageJson ={
                "text":message,
                "entitiesSub":message,
                "messageType":type,
                "dialogNumber":dialogNumber,
                "sendTime": datetime.datetime.now().strftime("%d/%m/%Y, %H:%M:%S")

            }

        self.context["messages"].append(messageJson)
        # update the tree
        self.contextTree =Tree(self.context)

    def findStuff(self,filter=None,stuff="intents"):
        matchedStuff=[]
        if filter:
            # $.intss[@.foo is 1][@.bar is b]
            query=["$.{0}".format(stuff)]
            for key in filter:
                if key != "confidence":
                    query.append("[@.{0} is {1}]".format(key,filter[key]))
                else:
                    query.append("[@.{0} > {1}]".format(key,filter[key]))

            query="".join(query)

            matchedStuff =list(self.contextTree.execute(query))
        return matchedStuff

    def findCurrentTopIntent(self):
        # Find current dialog number
        currentDialogNumber = self.context["dialogs"][-1]

        # Find
        #if confidenceLimit:
        topIntent= list(self.findStuff({"dialogNumber": currentDialogNumber, "rank": 1}, stuff="intents"))

        #print "topfound",topIntent
        if len(topIntent)>0:
            topIntent=topIntent[-1]
        else:
            topIntent={}

        return topIntent

    def findCurrentEntities(self):
        # Find current dialog number
        currentDialogNumber = self.context["dialogs"][-1]

        # Find current Entities
        currentEntities = list(self.findStuff({"dialogNumber": currentDialogNumber}, stuff="entities"))


        return currentEntities

    @staticmethod
    def findObjectInStuff(stuff,filter):

        allStuff={
            "allStuff":stuff,

        }
        stuffTree=Tree(allStuff)
        query=["$.allStuff"]
        for key in filter:
            query.append("[@.{0} is {1}]".format(key, filter[key]))

        query = "".join(query)
        return list(stuffTree.execute(query))






#from examples import entities as entitiesExtractorJson
import re
import json

class EntitiesHandler():
    """
    This class is responsible for handling entities within the bot
    The init method for the EntitiesHandler Class
        :param entitiesExtractorJson: this is a json which specifies the type of entities and their definitions . Note currently no schema checks have been implemented .

        **Sample Json accepted**

        .. code-block:: json


            {

                    "GreetingsHelper":{

                        "wsup" : [{
                                    "tag":"case-insensitive",
                                    "pattern":"\s[w]*[a]*[s]+[u]+[p]+\s",
                                    "type":"regex"


                        }
                                ],
                        "hi":[
                            {
                                "tag":"case-insensitive",
                                "pattern":"\s[h]+[i]+\s",
                                "type":"regex"


                            }

                        ],
                        "hello":[
                            {
                                "tag":"case-insensitive",
                                "pattern":"\s[h]+[e]+[l]+[o]+\s",
                                "type":"regex"


                            },
                            {
                               "tag":"case-insensitive",
                               "extractor":helloExtractorFunc ->
                               "type":"function"

                            }

                        ]


                    },

                    "LaughterHelper":{

                        "haha" : [{
                                    "tag":"case-insensitive",
                                    "pattern":"\s(h+(a|e)+)+(h+)?\s",
                                    "type":"regex"


                        }
                                ],
                        "happysmily":[
                            {
                                "tag":"case-insensitive",
                                "pattern":"\s\:\)\s",
                                "type":"regex"


                            }

                        ],

                    },
                    "CoolHelper":{
                        "cool":[
                            {
                                "tag":"case-insensitive",
                                "pattern":"\sc+oo+l+\s",
                                "type":"regex"

                            }


                        ]


                    },
                    "ByeHelper":{
                        "bye":[
                            {
                                "tag":"case-insensitive",
                                "pattern":"\s(goo+d)?b+y+e+\s",
                                "type":"regex"

                            }
                        ]

                    }

                }

        **kind** : Here GreetingsHelper,LaughterHelper,CoolHelper,ByeHelper are entity kinds.
        also "wsup","hi","hello" are entity values assosciated with the entity kind GreetingsHelper

        **type**: Only two types are supported -> regex and function.

        **pattern** : the pattern of the regex only used if type is regex,

        **extractor**: A function that extracts the given entity from the message text,

        Note that any function used should take in two arguments messageText and dialogNumber and should
        return A list of dicts of the format

        .. code-block:: json

            [{
                        "value":the value of the entity , for more details look at the sample json.,
                        "exactValue":the exact value that occurs within the message text.,
                        "kind":the kind of the entity eg  food in case of pizza.,
                        "dialogNumber":the dialog number of the current message -> same as the input param.,
                        "foundAt":[start location , end location],

            }]


        **substituter**: A custom substitution function to substitute entity by another value, This function takes in the message text and should
        return the substituted message text


    """

    def __init__(self,entitiesExtractorJson=None):
        if not entitiesExtractorJson:
            entitiesExtractorJson={}

        self.entitiesExtractorJson=entitiesExtractorJson


    def extractEntitySynonym(self,message,pattern,value,kind,tag,dialogNumber):
        """
            This function extracts the entities and synonyms from the text message and returns a json. Currently it does not extract any synonyms
                :param message: The text message str eg "I want to buy a pizza"
                :param pattern:
                :param value:
                :param kind:
                :param tag: can be "case-insensitive" if we want the extractor to ignore the casing (uppercase or lowercase) of the entity text
                :param dialogNumber: the dialog number of the message string.
                :return: A list of dicts of the format

                .. code-block:: json

                    [{
                                "value":the value of the entity , for more details look at the sample json.,
                                "exactValue":the exact value that occurs within the message text.,
                                "kind":the kind of the entity eg : food in case of pizza.,
                                "dialogNumber":the dialog number of the current message -> same as the input param.,
                                "foundAt":[start location , end location],

                    }]
        """


        entitiesFound=[]
        if "case-insensitive" in tag:
            iter=re.finditer(pattern=pattern,string= message,flags=re.I)
        else:
            iter = re.finditer(pattern=pattern, string=message)

        for entityDetected in iter:

            exactValue=message[entityDetected.start() : entityDetected.end()]
            entitiesFound.append(
                {
                    "value":value,
                    "exactValue":exactValue.strip(),
                    "kind":kind,
                    "dialogNumber":dialogNumber,
                    "foundAt":[entityDetected.start() , entityDetected.end()],
                    #"logic":synonymSpecs

                }
            )
        return entitiesFound


    def extractAllEntities(self,message,dialogNumber=0):
        """
        This function extracts all entities from the message .
            :param message: The message text
            :param dialogNumber:  the current dialog number
            :return: A list of dicts of the format

            .. code-block:: json

                [{
                            "value":the value of the entity , for more details look at the sample json.,
                            "exactValue":the exact value that occurs within the message text.,
                            "kind":the kind of the entity eg : food in case of pizza.,
                            "dialogNumber":the dialog number of the current message -> same as the input param.,
                            "foundAt":[start location , end location],

                }]
        """
        #message = " {0} ".format(message)
        allEntities=[]
        for entityKind in  self.entitiesExtractorJson.keys():
            for entityValue in self.entitiesExtractorJson[entityKind].keys():
                for synonymSpecs in self.entitiesExtractorJson[entityKind][entityValue]:
                    if synonymSpecs["type"] == "regex":
                        tag=synonymSpecs["tag"]
                        synonymPattern=synonymSpecs["pattern"]
                        entFound =  self.extractEntitySynonym(message,synonymPattern,entityValue,entityKind,tag,dialogNumber)
                    elif synonymSpecs["type"] == "function":
                        entFound = synonymSpecs["extractor"](message,dialogNumber)
                    else:
                        entFound =[]

                    allEntities.extend(entFound)


        return allEntities



    def substituteEntityKind(self,message,pattern,by,tag):
        """
        This function substitutes an entity value for its kind in the message text and returns the substituted message text .
        Note this function is currently unstable and a better version needs to be written to substitute entities effectively .
            :param message: The message text
            :param pattern: the entity pattern
            :param by:  the value to be substituted by eg if we need to substitute the word "equivalent" by "alike" then by ="alike"
            or if pizza then it will be kind
            :param tag:"case-insensitive" or not
            :return: the substituted message text
        """

        #message = " {0} ".format(message)
        if "case-insensitive" in tag:
            flags=re.I
        else:
            flags=None

        messageVar = re.sub( pattern , by, message,flags=flags)

        return messageVar

    def substituteAllEntities(self,message):
        """
        This function substitutes all entity values with entity kinds
            :param message:The message text
            :return: substituted message text
        """
        #message =" {0} ".format(message)
        for entityKind in  self.entitiesExtractorJson.keys():
            for entityValue in self.entitiesExtractorJson[entityKind].keys():
                for synonymSpecs in self.entitiesExtractorJson[entityKind][entityValue]:
                    if synonymSpecs["type"] == "regex":
                        tag=synonymSpecs["tag"]
                        synonymPattern=synonymSpecs["pattern"]
                        message = self.substituteEntityKind(message, synonymPattern, " --!{0} ".format(entityKind), tag)
                    elif synonymSpecs["type"] == "function":
                        message = synonymSpecs["substituter"](message)


        return message


















